#!/usr/bin/perl
#
# ACBLgamedecode.pm
#
# Matthew J. Kidd (San Diego, CA)
#
# This software is released under the GNU General Public License GPLv3
# See: http://www.gnu.org/licenses/gpl.html for full license.
#
# Partially decodes an ACBLscore game file, outputting result as JSON.
#
# 06-Feb-2015 - Current release
# 05-Nov-2014 - Initial release

package ACBLgamedecode;

use strict;

# Error codes
my $SUCCESS              =    0;
my $ERR_FILE_ERR         =   -2;
my $ERR_NEED_32_BIT_PERL =   -4;
my $ERR_NOT_GAME_FILE    = -257;

# Version of decoded data
my $DECODE_FORMAT_VERSION = 3;

# ACBLscore constants
my $MAX_EVENTS   =  50;
my $MAX_SECTIONS = 100;

# ACBLscore events types and scoring methods.
my @EVENT_TYPE = ('Pairs', 'Teams', 'Individual', 'Home Style Pairs', 'BAM', 'Series Winner');
my @EVENT_SCORING = ('Matchpoints', 'IMPs with computed datum', 'Average IMPs', 'Total IMPs',
  'Instant Matchpoints', 'BAM Teams', 'Win/Loss', 'Victory Points', 'Knockout', undef,
   'Series Winner', undef, undef, undef, undef, undef, 'BAM Matchpoints', undef, 'Compact KO');
my @MOVEMENT_TYPE = ('Mitchell', 'Howell', 'Web', 'External', 'External BAM', 'Barometer',
  'Manual Mitchell', 'Manual Howell');

my @CLUB_GAME_TYPE = ('Open', 'Invitational', 'Novice', 'BridgePlus', 'Pupil', 'Introductory');
my %ACBL_PLAYER_RANKS = (' ' => 'Rookie', 'A' => 'Junior Master', 'B' => 'Club Master', 
  'C' => 'Sectional Master', 'D' => 'Regional Master', 'E' => 'NABC Master',
  'F' => 'Advanced NABC Master', 'G' => 'Life Master', 'H' => 'Bronze Life Master', 
  'I' => 'Silver Life Master', 'J' => 'Gold Life Master', 'K' => 'Diamond Life Master',
  'L' => 'Emerald Life Master', 'M' => 'Platinum Life Master', 'N' => 'Grand Life Master'
);
my @AWARD_SETS = ('previous', 'current', 'total');
my $PIGMENTATION_TYPES = ' BSRGP';
my @RIBBON_COLORS = ('', 'Blue', 'Red', 'Silver', undef, undef, undef, undef, undef, 'Blue/Red');

my %SPECIAL_SCORES = ('900' => 'Late Play', '950' => 'Not Played',
  '2040' => 'Ave-', '2050' => 'Ave', '2060' => 'Ave+');
  
my $BYE_TEAM_NUMBER = 888;  

# Structure parameters
my $STRAT_STRUCTURE_SIZE = 95;
my $SECTION_SUMMARY_BASE = 0x13e;
my $SECTION_SUMMARY_SIZE = 22;
my $PLAYER_STRUCTURE_SIZE = 120;
my $TEAM_MATCH_ENTRY_SIZE = 32;

sub decode {
  # Loads and parses a game file
  if (scalar(@_) < 1 || scalar(@_) > 3) { die "function requires 1-3 arguments."; }
  my ($fname, $opt, $bws) = @_;
  if (ref($fname) ne '') { die '1st argument must be a scalar.'; }
  if (!defined $opt) { $opt = {}; }
  elsif (ref($opt) ne 'HASH') { die '2nd argument must be a hashref or undef.'; }
  if (defined $bws && ref($bws) ne 'HASH') { die "3rd argument must be a hashref or undef."}
  
  my $data;
  # Safely isolate $/ change.
  {
    local $/; my $fh;
    return($ERR_FILE_ERR) if !open($fh, '<', $fname);
    binmode($fh);
    $data = <$fh>;
    close($fh);
  }
  
  # Check for the 'AC3' magic bytes that appear at the start of all ACBLscore
  # game files.
  return($ERR_NOT_GAME_FILE) if (substr($data,0,6) ne "\x12\x0a\x03AC3");
  
  my ($p, $event_type_id, $event_scoring_id, $rankstr, $pMemoNote);

  my $gm = {'filename' => $fname, 'decode_format_version' => $DECODE_FORMAT_VERSION };
  $gm->{'ACBLscore_version'} = sprintf('%.2f', unpack('v', substr($data, 0x9db, 2)) / 100);
  $gm->{'ACBLscore_min_compatible_version'} = sprintf('%.2f', unpack('v', substr($data, 0x9e1, 2)) / 100);
  $gm->{'creation_timestamp'} = datetime($data, 0x9dd);
  $pMemoNote = unpack('V', substr($data, 0x9d6, 4));
  $gm->{'memo'} = memonote($data, $pMemoNote, 'memo') if $pMemoNote;
  $pMemoNote = unpack('V', substr($data, 0x9e3, 4));
  $gm->{'note'} = memonote($data, $pMemoNote, 'note') if $pMemoNote;
  
  my $ix = 0;
  for (my $i=0; $i<$MAX_EVENTS; $i++) {
    $p = unpack('V', substr($data, 0x12 + 4*$i, 4));
    next if !$p;
    $event_type_id    = unpack('C', substr($data, 0xda  + $i));
    $event_scoring_id = unpack('C', substr($data, 0x10c + $i));
    my $ev = {'event_id' => $i+1,  
      'event_type_id' => $event_type_id, 'event_type', $EVENT_TYPE[$event_type_id], 
      'event_scoring_id' => $event_scoring_id, 'event_scoring', $EVENT_SCORING[$event_scoring_id] };
      
    $rankstr = eventDetails($ev, $data, $p);
    addSectionCombining($ev, $data);
    addSections($ev, $data, $opt, $rankstr, $bws);
    $gm->{'event'}[$ix] = $ev;
    $ix++;
  }
  
  return ($SUCCESS, $gm);
}

sub zstring {
  # zstring(str, offset)
  #
  # Parse a string where first byte gives the length of the string. 

  return substr($_[0], $_[1]+1, unpack('C', substr($_[0], $_[1],1)));
}

sub zstrings {
  # Decode multiple zstring fields and add fields to a hashref.
  # Example: zstrings($hr, $data, $p, 'fdname1', offset1, ...)
  die "At least two arguments are required" if (scalar(@_) < 3);
  die "Last string field has no offset" if (scalar(@_) % 2 == 0);
  my $hr = shift; my $data = shift; my $p = shift; my $fdname;
  while (defined ($fdname = shift)) { $hr->{$fdname} = zstring($data, $p + shift); }
}

sub datetime {
  # Decode ACBLscore (Pascal?) 32-bit encoded date and time.
  my ($data, $p) = @_;
  my ($tm, $dt) = unpack('vv', substr($data, $p, 4));
  my $year  = ($dt >> 9) + 1980;
  my $month = ($dt >> 5) & 0x0F;
  my $day   = $dt & 0x0F;
  my $hour  = $tm >> 11;
  my $min   = ($tm >> 5) & 0x3F;
  my $sec   = ($tm << 1) & 0x1F;
  
  return sprintf('%d-%02d-%02dT%02d:%02d:%02d', $year, $month, $day, $hour, $min, $sec);
}

sub memonote {
  # Decodes an ACBLscore Memo or Note structure.
  my ($data, $p, $type) = @_;
  my $linelen = $type eq 'memo' ? 48 : $type eq 'note' ? 76 : 0;
  die '3rd argument must be "memo" or "note"' if ! $linelen;
  my $linecount = unpack('v', substr($data, $p + 4, 2));
  my $text = '';
  for (my $i=0; $i < $linecount; $i++) {
    $text .= zstring($data, $p + 6 + $i * $linelen) . "\n";
  }
  
  return $text;
}

sub eventDetails {
  # Parse event details.
  if (scalar(@_) != 3) { die "function requires 3 arguments."; }
  my ($ev, $data, $p) = @_;

  my $isTeams = $ev->{'event_type_id'} == 1;
  
  $ev->{'mp_rating'} = {
    'p-factor' => unpack('v', substr($data, $p + 0x7d)) / 1000,
    't-factor' => unpack('v', substr($data, $p + 0x83)) / 1000,
    's-factor' => unpack('v', substr($data, $p + 0x24f)) / 100
  };
 
  my $club_session_num = unpack('C', substr($data, $p + 0x95));
  # It is not clear the best way to figure out if results are from a tournament.
  # One way is to check whether there is a club number. Another way is to look
  # at the club session number which is what is done below.
  my $isTourney = $club_session_num == 0 ? 1 : 0;
 
  $ev->{'tournament_flag'} = $isTourney;
  if (! $isTourney) {
    $ev->{'club_num'} = zstring($data, $p + 0xb0);
    $ev->{'club_session_num'} = $club_session_num;
    $ev->{'club_game_type'} = $CLUB_GAME_TYPE[unpack('C', substr($data, $p + 0xa1, 1))];
  }
  $ev->{'session_num'} = unpack('C', substr($data, $p + 0x8c, 1));
  $ev->{'nstrats'} = unpack('C', substr($data, $p + 0x9e, 1));
  $ev->{'nsessions'} = unpack('C', substr($data, $p + 0x9f, 1));
  $ev->{'consolation_flag'} = unpack('C', substr($data, $p + 0xa0, 1));
  $ev->{'modification_timestamp'} = datetime($data, $p + 0xb7);  
  $ev->{'nbrackets'} = unpack('C', substr($data, $p + 0xc2, 1)) if $isTeams;
  $ev->{'bracket_num'} = unpack('C', substr($data, $p + 0xc3, 1)) if $isTeams;
  $ev->{'continuous_pairs_flag'} = unpack('C', substr($data, $p + 0xce, 1)) if ! $isTeams;
  $ev->{'senior_event_flag'} = unpack('C', substr($data, $p + 0x251, 1));
  $ev->{'mp_award_revision_num'} = unpack('C', substr($data, $p + 0x254, 1));
  $ev->{'stratify_by_avg_flag'} = unpack('C', substr($data, $p + 0x2c8, 1));
  $ev->{'non_ACBL_flag'} = unpack('C', substr($data, $p + 0x2ca, 1));
  # The +0 forces Perl's internal representation to be numeric instead of a string.
  # It doesn't matter for Perl but it does alter the JSON output, e.g. 0 instead of "0"
  $ev->{'final_session_flag'} = ($ev->{'session_num'} == $ev->{'nsessions'}) + 0;
  
  my $club_or_tournament = $isTourney ? 'tournament' : 'club';
  my $director_or_city   = $isTourney ? 'city' : 'director';
  zstrings($ev, $data, $p, 'event_name', 0x4, 'session_name', 0x1e, 
    $director_or_city, 0x2c, 'sanction', 0x3d, 'date', 0x48, $club_or_tournament, 0x5c,
    'event_code', 0x76, 'qual_event_code', 0xc5, 'hand_set', 0x244);

  my $pMemo = unpack('V', substr($data, $p + 0xbd, 4));
  $ev->{'memo'} = memonote($data, $pMemo, 'memo') if $pMemo;

  my $rankstr;
  for (my $i=0; $i<$ev->{'nstrats'}; $i++) {
    $ev->{'strat'}[$i] = strat($data, $p + 0xd4 + $i * $STRAT_STRUCTURE_SIZE);
    $rankstr .=  $ev->{'strat'}[$i]{'letter'};
  }
  
  return $rankstr;
}

sub strat {
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($data, $p) = @_;
  
  my $st = {
    'first_overall_award' => unpack('v', substr($data, $p + 0x10, 2)) / 100,
    'ribbon_color' => $RIBBON_COLORS[unpack('C', substr($data, $p + 0x12, 1))],
    'ribbon_depth' => unpack('C', substr($data, $p + 0x13, 1)),
    'mpt_factor' => unpack('V', substr($data, $p + 0x14, 4)) / 10000,
    'overall_award_depth' => unpack('v', substr($data, $p + 0x18, 2)),
    'table_basis' => unpack('C', substr($data, $p + 0x1a, 2)),
    'min_mp' => unpack('v', substr($data, $p + 0x1e, 2)),
    'max_mp' => unpack('v', substr($data, $p + 0x20, 2)),
    'letter' => substr($data, $p + 0x22, 1),
    'club_pct_open_rating' => unpack('C', substr($data, $p + 0x23, 1)),
    'pigmentation_breakdown' => {
      'overall' => pigmentation($data, $p + 0x32),
      'session' => pigmentation($data, $p + 0x41),
      'section' => pigmentation($data, $p + 0x50)
    }
  };

}

sub pigmentation {
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($data, $p) = @_;
  
  my (@pgset, $i, $j, $mp, $tp);
  my @pct = unpack('v3', substr($data, $p, 6));
  my @mp  = unpack('v3', substr($data, $p+6, 6));
  my @tp  = unpack('C3', substr($data, $p+12, 12));
  
  for ($i=0; $i<3; $i++) {
    last if ! $pct[$i];
    push @pgset, {'pct' => $pct[$i] / 100, 'mp' => $mp[$i] / 100,
      'type' => substr($PIGMENTATION_TYPES, $tp[$i], 1) }; 
  }
  return \@pgset;
}

sub addSectionCombining {
  # Work out how sections are combined and ranked together. Only 
  # combined sections may optionally be ranked together.
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($ev, $data) = @_;
  
  my ($prevScore, $prevRank, $nextScore, $nextRank);
  my @combining_and_ranking;

  # Search Master Table for all sections belonging to the event.
  my ($p, $pc, $pr);
  for (my $i=0; $i<$MAX_SECTIONS; $i++) {
    $p = $SECTION_SUMMARY_BASE + $SECTION_SUMMARY_SIZE * $i;
    next if unpack('C', substr($data, $p, 1)) != $ev->{'event_id'};
    
    $prevScore = unpack('C', substr($data, $p + 0xf, 1));
    next if $prevScore != 0;
    # Found first section in a group of combined sections.
    my @combined;
    $pc = $p;
    while (1) {
      $prevRank = unpack('C', substr($data, $pc + 0x11, 1));
      if ($prevRank == 0) {
        # Found first section in a group of sections ranked together (applies to
        # section awards, i.e. N-S and E-W can be ranked across multiple sections).
        my @rankedTogether;
        $pr = $pc;
        while (1) {
          push @rankedTogether, zstring($data, $pr + 0x1);
          $nextRank = unpack('C', substr($data, $pr + 0x12));
          if ($nextRank == 0) { push @combined, \@rankedTogether; last; }
          $pr = $SECTION_SUMMARY_BASE + $SECTION_SUMMARY_SIZE * ($nextRank-1);
        }
      }
      $nextScore = unpack('C', substr($data, $pc + 0x10, 1));
      last if $nextScore == 0;
      $pc = $SECTION_SUMMARY_BASE + $SECTION_SUMMARY_SIZE * ($nextScore-1);      
    }
    push @combining_and_ranking, \@combined;
  }
  
  $ev->{'combining_and_ranking'} = \@combining_and_ranking;
}

sub addSections {
  if (scalar(@_) != 5) { die "function requires 5 arguments."; }
  my ($ev, $data, $opt, $rankstr, $bws) = @_;
  
  my $event_type_id = $ev->{'event_type_id'};
  
  # Search Master Table for all sections belonging to the event.
  my $p = $SECTION_SUMMARY_BASE;
  for (my $i=0; $i<$MAX_SECTIONS; $i++) {
    $p = $SECTION_SUMMARY_BASE + $i * $SECTION_SUMMARY_SIZE;
    next if unpack('C', substr($data, $p, 1)) != $ev->{'event_id'};
    my $sc = section($data, $p, $opt, $event_type_id, $rankstr, $bws);
    $ev->{'section'}{$sc->{'letter'}} = $sc; 
  }
}

sub section {
  if (scalar(@_) != 6) { die "function requires 6 arguments."; }
  my ($data, $p, $opt, $event_type_id, $rankstr, $bws) = @_;
  
  my $sc = {'letter' => zstring($data, $p+1), 'rounds' => unpack('C', substr($data, $p + 0x14)) };
  $sc->{'is_scored'} = unpack('C', substr($data, $p + 0xe)) & 1;

  # Board Results pointer
  my $pBoardIdx = unpack('V', substr($data, $p + 8));
  
  # Move on to Section Details Structure.
  $p = unpack('V', substr($data, $p + 4));

  my @pIndex = unpack('V4', substr($data, $p + 4));
  my $isTeams = $event_type_id == 1;
  my $isIndy  = $event_type_id == 2;
  my $isBAM   = 0 + ($event_type_id == 4);
  
  my $highest_pairnum = unpack('v', substr($data, $p + 0x1b));
  
  $sc->{'movement_type'} = $MOVEMENT_TYPE[ unpack('C', substr($data, $p + 0x18)) ] if ! $isTeams;
  $sc->{'is_barometer'} = unpack('C', substr($data, $p + 0x47)) if ! $isTeams;
  $sc->{'is_web'} = unpack('C', substr($data, $p + 0x60)) if ! $isTeams;
  $sc->{'is_bam'} = $isBAM;
  
  $sc->{'nboards'} = unpack('v', substr($data, $p + 0x19)) if ! $isTeams;
  $sc->{$isBAM ? 'highest_teamnum' : 'highest_pairnum'} = $highest_pairnum if ! $isTeams;
  $sc->{'boards_per_round'} = unpack('C', substr($data, $p + 0x1d));
  $sc->{'max_results_per_board'} = unpack('C', substr($data, $p + 0x61)) if ! $isTeams;
  $sc->{'board_top'} = unpack('v', substr($data, $p + 0x1e)) if ! $isTeams;
  $sc->{'ntables'} = unpack('v', substr($data, $p + 0x48));
  
  $sc->{'match_award'} = unpack('v', substr($data, $p + 0xb5)) / 100 if $isTeams;
  $sc->{'maximum_score'} = unpack('v', substr($data, $p + 0x4e));
  $sc->{'modification_timestamp'} = datetime($data, $p + 0xbd);

  my $pMemo = unpack('V', substr($data, $p + 0xd1, 4));
  $sc->{'memo'} = memonote($data, $pMemo, 'memo') if $pMemo;

  return $sc if $opt->{'sectionsonly'};

  my $isHowell = unpack('C', substr($data, $p + 0x18)) == 1 ? 1 : 0;
  my $phantom = unpack('c', substr($data, $p + 0x43));
  ## print $sc->{'highest_pairnum'}, "\n";
  
  my $bws2 = defined $bws ? $bws->{$sc->{'letter'}} : undef;
  
  my ($i, $j);
  if ($isHowell) {
    # Howell can be a pairs or an individual event.
    $sc->{'is_howell'} = $isHowell;
    if ( !$opt->{'noentries'} ) {    
      my $pPNM = $p + 0xdd;
      my ($pairnum, $table, $dir, $str, $pEntry);
      my @reassign = unpack('C80', substr($data, $p + 0xdd, 80));
      for ($i=1; $i<=$highest_pairnum; $i++) {
        $pairnum = $i;
        next if $pairnum == $phantom;
        $str = substr($data, $pPNM + 0x50 + 2*($pairnum-1), 2);
        ($table, $dir) = unpack('C2', $str);
        $pEntry = unpack('V', substr($data, $pIndex[$dir-1] + 0x10 + 8 * $table, 4));
  
        # Rare pair number reassignments with ACBLscore EDMOV command.
        $pairnum = $reassign[$i-1] if $reassign[$i-1];
  
        # IF clause is a quick hack to filter out phantom pairs.
        $sc->{'entry'}{$pairnum} = entry($data, $pEntry, $opt, $rankstr)
          if unpack('v', substr($data, $pEntry + 0x1c, 2)) != 0;
      }
    }

    $sc->{'board'} = boards($data, $pBoardIdx, $p, $isIndy, $bws2) if !$opt->{'noboards'};
  }
  elsif (!$isTeams && !$isBAM) {
    # Mitchell movement
    $sc->{'is_howell'} = $isHowell;
    if ( !$opt->{'noentries'} ) {
    my $pPNM = $p + 0xdd;
      my $ndir = $isIndy ? 4 : 2;
      my ($pairnum, $table, $pEntry);
      my @dirletter = ('N', 'E', 'S', 'W'); 
      my @reassign = unpack('C160', substr($data, $p + 0xdd, 160));
      for ($i=1; $i<=$highest_pairnum; $i++) {
        for ($j=0; $j<$ndir; $j++) {
          $pairnum = $i;
          next if $pairnum == $phantom && $j == 0 || $pairnum == -$phantom && $j == 1;
          $table = unpack('C', substr($data, $pPNM + 0xa0 + 4*($pairnum-1) + $j, 1) );
          $pEntry = unpack('V', substr($data, $pIndex[$j] + 0x10 + 8 * $table, 4));
  
          # Rare pair number reassignments with ACBLscore EDMOV command.
          $pairnum = $reassign[4*($i-1)+$j] if $reassign[4*($i-1)+$j];
          
          # IF clause is a quick hack to filter out phantom pairs.
          $sc->{'entry'}{$pairnum . $dirletter[$j]} = entry($data, $pEntry, $opt, $rankstr)
            if unpack('v', substr($data, $pEntry + 0x1c, 2)) != 0;
        }
      }
    }

    $sc->{'board'} = boards($data, $pBoardIdx, $p, $isIndy, $bws2) if !$opt->{'noboards'};
  }
  elsif ($isTeams || $isBAM) {
    if ( !$opt->{'noentries'} ) {    
      my ($teamnum, $nplayers, $pTeam, $pEntry);
      my $nteams = unpack('C', substr($data, $pIndex[0] + 6, 1));
      $sc->{'nteams'} = $nteams;
      
      $pTeam = $pIndex[0] + 0x14;
      for ($i=0; $i<$nteams; $i++) {
        $teamnum  = unpack('v', substr($data, $pTeam, 2));
        $nplayers = unpack('v', substr($data, $pTeam + 2, 2));
        $pEntry   = unpack('V', substr($data, $pTeam + 4, 4));
        $sc->{'entry'}{$teamnum} = entry($data, $pEntry, $opt, $rankstr, $nplayers);
        $pTeam += 8;
      }
    }
    
    if ($isTeams) {
      my $pTeamMatch = unpack('V', substr($data, $p + 0x23d));
      $sc->{'matches'} = teamMatches($data, $pTeamMatch, $pIndex[0]) if $pTeamMatch;
    }
    
    if ($isBAM) {
      $sc->{'board'} = boards($data, $pBoardIdx, $p, $isIndy, $bws2) if !$opt->{'noboards'};
    }
  }

  return $sc;
}

sub entry {
  if (scalar(@_) != 4 && scalar(@_) != 5) { die "function requires 3-4 arguments."; }
  my ($data, $p, $opt, $rankstr, $award, $nplayers) = @_;
  
  my $en;
  my @intfloats = unpack('l<6', substr($data, $p + 0x4, 24));
  
  $en->{'score_adjustment'} = $intfloats[0] / 100;
  $en->{'score_unscaled'}   = $intfloats[1] / 100;
  $en->{'score_session'}    = $intfloats[2] != -1 ? $intfloats[2] / 100 : undef;
  $en->{'score_carrover'}   = $intfloats[3] / 100;
  $en->{'score_final'}      = $intfloats[4] != -1 ? $intfloats[4] / 100 : undef;
  $en->{'score_handicap'}   = $intfloats[5] / 100;
  
  $en->{'pct'} =  unpack('v', substr($data, $p + 0x1c, 2)) / 100;
  $en->{'strat_num'} =  unpack('C', substr($data, $p + 0x1e, 1));
  $en->{'mp_average'} =  unpack('v', substr($data, $p + 0x20, 2));
  $en->{'nboards'} =  unpack('C', substr($data, $p + 0x2f, 1));
  $en->{'eligibility'} =  unpack('C', substr($data, $p + 0x33, 1));

  $award = award($data, $p + 0x34, $rankstr);
  $en->{'award'} = $award if defined $award;
  $en->{'rank'} = rank($data, $p + 0x5e);
  
  # There are three ways to determine the maximum number of players in an entry 
  # structure. The method here is based on the size of the player structure and will
  # return 1, 2, or 6. The second and probably more proper method is to use the number
  # from the entry index table. The third is to infer it from the event type.
  $nplayers = (unpack('v', substr($data, $p + 0, 2)) - 0xa2) / $PLAYER_STRUCTURE_SIZE
    if !defined $nplayers;
  
  if ($nplayers <= 2) {
    for (my $i=0; $i<$nplayers; $i++) {
      $en->{'player'}[$i] = player($data, $p + 0xa4 + $i * $PLAYER_STRUCTURE_SIZE, $opt, $rankstr);    
    }
  }
  else {
    # Typically only have four players on a team but up to six are possible.
    # Only return the player structures for actual players.
    my $j=0;
    for (my $i=0; $i<$nplayers; $i++) {
      my $pl = player($data, $p + 0xa4 + $i * $PLAYER_STRUCTURE_SIZE, $opt, $rankstr);
      next if $pl->{'lname'} eq '' && $pl->{'fname'} eq '' && $pl->{'pnum'} eq '';
      $en->{'player'}[$j] = $pl;
      $j++ 
    }
  }
  
  return $en;
}

sub player {
  if (scalar(@_) != 4) { die "function requires 3 arguments."; }
  my ($data, $p, $opt, $rankstr) = @_;
  
  my ($pl, $award);
  $pl->{'team_wins'} = unpack('v', substr($data, $p + 0x44, 2)) / 100;
  $pl->{'mp_total'}  = unpack('v', substr($data, $p + 0x71, 2));
  $pl->{'acbl_rank_letter'} = substr($data, $p + 0x73, 1);
  $pl->{'acbl_rank'} = $ACBL_PLAYER_RANKS{$pl->{'acbl_rank_letter'}};

  zstrings($pl, $data, $p, 'lname', 0, 'fname', 0x11, 'city', 0x22, 'state', 0x33,
    'pnum', 0x36, 'country', 0x75);
  $pl->{'db_key'} = zstring($data, $p + 0x3e) if $opt->{'binaryfields'};
    
  $award = award($data, $p + 0x48, $rankstr);
  $pl->{'award'} = $award if defined $award;

  return $pl;
}

sub award {
  if (scalar(@_) != 3) { die "function requires 3 arguments."; }
  my ($data, $p, $rankstr) = @_;

  my ($awset, $i, $j, @v, $anyAward, $reason);
  for ($i=0; $i<3; $i++) {
    @v = unpack('(vCC)3', substr($data, $p + $i * 12, 12) );
    my @aw;
    for ($j=0; $j<=6; $j+=3) {
      last if ! $v[$j];
      $anyAward = 1;
      $reason = $v[$j+2] ?
        ($v[$j+2] >= 10 ? 'S' : 'O') . substr($rankstr, $v[$j+2] % 10 - 1, 1) : '';
      push @aw, [$v[$j] / 100, substr($PIGMENTATION_TYPES, $v[$j+1], 1), $reason, $v[$j+2]];
    }
    $awset->{ $AWARD_SETS[$i] } = \@aw;
  }
  return $anyAward ? $awset : undef;
}

sub rank {
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($data, $p) = @_;
  
  my (@rkset, $i, @v);
  for ($i=0; $i<3; $i++) {
    # Currently not unpacking pointers to next lowest rank.
    @v = unpack('v6', substr($data, $p + $i * 20, 20));
    # Don't include zeros from strats that entry or player is not eligible to
    # be ranked it.
    last if !$v[5];
    my $rk = {'section_rank_low' => $v[0], 'section_rank_high' => $v[1],
      'overall_rank_low' => $v[2], 'overall_rank_high' => $v[3],
      'qual_flag' => $v[4], 'rank' => $v[5] };
    push @rkset, $rk;
  }
  return \@rkset;
}

sub boards {
  if (scalar(@_) != 5) { die "function requires 5 arguments."; }
  my ($data, $pBoardIdx, $pSection, $isIndy, $bws2) = @_; 
  
  my (%bdset, $i, $j, $k, $bnum, $p, $nresults, $dr);
  my $ncompetitors = $isIndy ? 4 : 2;
  my $fmt = 'CC(vs<V)' . $ncompetitors;
  my $resultSize = 2 + 8 * $ncompetitors;
  my $nboards = unpack('v', substr($data, $pBoardIdx + 4, 4));
  my $kupper = 3 * $ncompetitors + 2;

  # Deal with possible EDMOV pair reassignment, first checking if any reassignments
  # have occurred to minimize time spent in the loops below.
  my $isHowell = unpack('C', substr($data, $pSection + 0x18));
  my @reassign = $isHowell ? unpack('C80', substr($data, $pSection + 0xdd, 80)) :
    unpack('C160', substr($data, $pSection + 0xdd, 160));
  my $anyReassignedPairs = 0;
  my ($bws3, $contract, $contractresult, $declarer, $lead);
  # It's not clear if individual reassignments are possible in an individual event.
  # Don't both for such. Individual events are rare and reassignments are rarer.
  if (!$isIndy) { foreach my $val (@reassign) { if ($val) {$anyReassignedPairs = 1; last; } } }

  for ($i=0; $i<$nboards; $i++) {
    my @bd;
    ($bnum, $nresults, $p) = unpack('vvV', substr($data, $pBoardIdx + 0x26 + $i * 8, 8));
    $bws3 = $bws2->{$bnum}; 

    $p += 6;
    for ($j=0; $j<$nresults; $j++) {
      my @v = unpack($fmt, substr($data, $p + $j * $resultSize, $resultSize));
      # Skip if board is not in play on this round.
      next if $v[3] == 999;
      for ($k=2; $k<$kupper; $k+=3) {
        if ($anyReassignedPairs) {
          if ($isHowell) { $v[$k] = $reassign[$v[$k]-1] if $reassign[$v[$k]-1] }
          else {
            # 0 for N-S, 1 for E-W
            $dr = ($k == 5);
            $v[$k] = $reassign[4*($v[$k]-1)+$dr] if $reassign[4*($v[$k]-1)+$dr];
          }  
        }
        if ($v[$k+1] < 900) { $v[$k+1] *= 10; }
        else {
          $v[$k+1] = $SPECIAL_SCORES{$v[$k+1]};
          $v[$k+1] = 'Unknown' if !defined $v[$k+1];
        }
        $v[$k+2] /= 100;
      }
    
      # BWS2 has electronic scoring results for the current section (if available).
      if (defined $bws3) {
        # $v[2] has the N-S pair number.
        $contract = $bws3->{$v[2]}{'Contract'};
        if (!$contract) {
          # Typically have an empty string for $CONTRACT when the board is not played
          # because the pair was too slow. Can also have an empty string if pairs are
          # assigned AVE, AVE+, or AVE-.
          push @v, '', '', '', '', undef;
        }
        elsif ($contract eq 'PASS') {
          push @v, 'PASS', '', '', '', undef;
        }
        else {
          substr($contract,2,3) = 'N' if substr($contract,2,2) eq 'NT';
          $contract =~ tr/x/*/; $contract =~ s/\s//g;
          push @v, 0 + substr($contract,0,1), substr($contract,1);
          
          $contractresult = $bws3->{$v[2]}{'Result'};
          $contractresult = 0 if $contractresult eq '=';
          $declarer = $bws3->{$v[2]}{'NS/EW'};
          push @v, $declarer, 0 + $contractresult, ;
          
          $lead = $bws3->{$v[2]}{'LeadCard'};
          push @v, (defined $lead ? $lead : '');
        }
      }
          
      push @bd, \@v;
    }
    $bdset{$bnum} = \@bd;
  }
  return \%bdset;
}

sub teamMatches {
  if (scalar(@_) != 3) { die "function requires 3 arguments."; }
  my ($data, $p, $pIndex) = @_;
  
  my ($i, $j, $nmatches, $pTMT, $pTME, $vsTeamID);
  
  # Construct mapping from Team Entry ID to Team number.
  my ($nteams, $nrounds) = unpack('vv', substr($data, $p + 4, 4));
  my @tmap = unpack("(vx6)$nteams", substr($data, $pIndex + 0x14, 8 * $nteams));
  my @pTMT = unpack("V$nteams", substr($data, $p + 0x56, 4 * $nteams));
  
  my $tmset;
  for ($i=0; $i<$nteams; $i++) {
    my @tm;
    # Pointer to Team Match Table
    $pTMT = $pTMT[$i];
    # A team might not play all rounds of the event, e.g. when a team gets
    # knocked out of Knockout event.
    $nmatches = unpack('C', substr($data, $pTMT + 4, 1));
    for ($j=0; $j<$nmatches; $j++) {
      # Team Match Entry
      $pTME = $pTMT + 0x22 + $j * $TEAM_MATCH_ENTRY_SIZE;
      $vsTeamID = unpack('v', substr($data, $pTME + 2, 2));
      push @tm, {
        'round' => unpack('C', substr($data, $pTME + 1, 1)),
        'vs_team' => unpack('v', substr($data, $pTME + 2, 2)),
        'IMPs' => unpack('s<', substr($data, $pTME + 8, 2)),
        'VPs' => unpack('v', substr($data, $pTME + 0x0a, 2)) / 100,
        'nboards' => unpack('C', substr($data, $pTME + 0x0d, 1)),
        'wins' => unpack('v', substr($data, $pTME + 0x16, 2)) / 100
      } 
    }
    $tmset->{$tmap[$i]} = \@tm;
  }
  
  return $tmset;
}

sub readBWSfiles {
  # Read Bridgemate / BridgePad / BridgeScorer BWS file(s).
  #
  # Input is an arrayref of filenames or a single filename as a scalar.  
  if (scalar(@_) != 1) { die "function requires 1 argument."; }
  my ($fnames) = @_;
  if ( ref($fnames) eq '' ) { $fnames = [$fnames]; }
  elsif ( ref($fnames) ne 'ARRAY') { 
    error('First argument must be an arrayref or a scalar');
  }

  my ($bws, $errmsg);
  foreach my $fname (@$fnames) {
    if (!-f $fname) { return "File does not exist: $fname"; }
  
    my ($bsc, $brd);
    my $ctrcnt = 0; my $leadcnt = 0;
    
    if ($^O eq 'MSWin32') {
      require DBI;
      require DBD::ODBC;
      
      my $dsn = "driver=Microsoft Access Driver (*.mdb);dbq=$fname";
      my $dbh = DBI->connect("dbi:ODBC:$dsn", {ReadOnly => 1, PrintError => 0});
      
      if (!$dbh) { return "Database open for $fname failed. Was it a BWS file?\n" . DBI->errstr; }
      
      my $sql;
      $sql = "select * from Section";
      $bsc = $dbh->selectall_hashref($sql, ['ID']);
      return "Database query failed: $sql\n" . $dbh->errstr if $dbh->err;
           
      $sql = "select * from ReceivedData";
      $brd = $dbh->selectall_hashref($sql, ['Section', 'Board', 'PairNS']);
      return "Database query failed: $sql\n" . $dbh->errstr if $dbh->err;  
          
      $dbh->disconnect;
    }
    else {
      # Rely on mdb-export program from mdbtools package on non-Windows platforms.
      my ($data, $colnames, $rownum, $row);
      ($errmsg, $data, $colnames) = mdbReadTable($fname, 'Section');
      return $errmsg if $errmsg;
      
      while ( ($rownum, $row) = each %$data ) { $bsc->{$row->{'ID'}} = $row; }
      
      ($errmsg, $data, $colnames) = mdbReadTable($fname, 'ReceivedData');
      return $errmsg if $errmsg;
      
      while ( ($rownum, $row) = each %$data ) {
        $brd->{$row->{'Section'}}{$row->{'Board'}}{$row->{'PairNS'}} = $row;
        $ctrcnt++ if $row->{'Contract'} ne '';
        $leadcnt++ if $row->{'LeadCard'} ne '';
      }
    }
  
    foreach my $id (sort { $bsc->{$a}{'Letter'} cmp $bsc->{$b}{'Letter'} } keys %$bsc ) {
      my $secletter = $bsc->{$id}{'Letter'}; 
      
      # Strip whitespace from section letters. Problems usually seems to be one
      # or two trailing spaces but try to guard against other cases too.
      $secletter =~ s/\s+//g;
      
      # Transfer sections to $BWS, noting any replacements.
      if ( exists $bws->{$secletter} ) {
        print STDERR "Previous section $secletter data replaced by results in file: $fname\n";
      }
      
      $bws->{$secletter} = $brd->{$id};
    }
  }
  
      
  # BridgePad stores the opening lead as value-suit instead of suit-value as
  # Bridgemate does. Convert BridgePad format to Bridgemate format for standard
  # handling.
  my ($section, $bnum, $NSPair, $row, $lead);
  foreach $section (keys %$bws) {
    foreach $bnum (keys %{$bws->{$section}} ) {
      while ( ($NSPair, $row) = each %{$bws->{$section}{$bnum}} ) {
        $lead = $row->{'LeadCard'};
        if (defined $lead && substr($lead,-1) =~ /[CDHS]/) {
          $bws->{$section}{$bnum}{$NSPair}{'LeadCard'} = substr($lead,-1) . substr($lead,0,-1);
        }
      }
    }
  }

  return('', $bws);
}

sub mdbReadTable {
  # Read a table from a JET database using mdb-export program of mdbtools package.
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($fname, $tname) = @_;
  
  my $mdbname = 'mdb-export';
  
  my $cmd = "\"$mdbname\" -Q \"$fname\" \"$tname\"";
  my $str = `$cmd`;
  if (${^CHILD_ERROR_NATIVE} != 0) {
    print <<"END";

  Failed to read \"$tname\" table from file \"$fname\" using mdb-export.
  The exit code was ${^CHILD_ERROR_NATIVE}.

  Maybe the mdbtools package is not installed on your system. On Ubuntu
  (Debian) you can install it via:

    apt-get install mdbtools       (if logged in as root)
    sudo apt-get install mdbtools  (if you have permission to install packages)
    
  Prepackaged binaries and dynamic libraries are available for Mac OS X.
  See the online documentation for ACBLgamedump.pl
  
END

    return("Failed to read \"$tname\" table from file \"$fname\" using mdb-export.");
  }
  
  my $rownum = 0;
  my (%data, @fd);
  open(my $fh, '<', \$str);
  my $fline = <$fh>;
  chomp($fline);
  my @colnames = split /,/, $fline;
  while (defined ($fline = <$fh>)) {
    chomp($fline);
    @fd = split /,/, $fline;
    $rownum++;
    for (my $i=0; $i<$#colnames; $i++) { $data{$rownum}{$colnames[$i]} = $fd[$i]; }
  }
  close($fh);

  return ('', \%data, \@colnames);
}

# This is required to inform Perl that package has been loaded successfully.
1;
